@@ -5,7 +5,7 @@ import signal
 signal.signal(signal.SIGINT, signal.SIG_DFL)
 
 import sys, os, datetime, time, copy, json, traceback, csv, string
-import multiprocessing, multiprocessing.queues, subprocess, re
+import multiprocessing, multiprocessing.queues, subprocess, re, ctypes
 from optparse import OptionParser
 
 try:
@@ -17,8 +17,8 @@ except ImportError:
 
 info = "'rethinkdb export` exports data from a RethinkDB cluster into a directory"
 usage = "\
-  rethinkdb export [-c HOST:PORT] [-a AUTH_KEY] [-f (csv | json)] [-d DIR]\n\
-      [--fields FIELD,FIELD...] [-e (DB | DB.TABLE)]..."
+  rethinkdb export [-c HOST:PORT] [-a AUTH_KEY] [-d DIR] [-e (DB | DB.TABLE)]...\n\
+      [--format (csv | json)] [--fields FIELD,FIELD...] [--clients NUM]"
 
 def print_export_help():
     print info
@@ -35,6 +35,8 @@ def print_export_help():
     print "                                   (required for CSV format)"
     print "  -e [ --export ] (DB | DB.TABLE)  limit dump to the given database or table (may"
     print "                                   be specified multiple times)"
+    print "  --clients NUM                    number of tables to export simultaneously (defaults"
+    print "                                   to 3)"
     print ""
     print "EXAMPLES:"
     print "rethinkdb export -c mnemosyne:39500"
@@ -60,6 +62,7 @@ def parse_options():
     parser.add_option("-d", "--directory", dest="directory", metavar="DIRECTORY", default=None, type="string")
     parser.add_option("-e", "--export", dest="tables", metavar="DB | DB.TABLE", default=[], action="append", type="string")
     parser.add_option("--fields", dest="fields", metavar="<FIELD>,<FIELD>...", default=None, type="string")
+    parser.add_option("--clients", dest="clients", metavar="NUM", default=3, type="int")
     parser.add_option("-h", "--help", dest="help", default=False, action="store_true")
     (options, args) = parser.parse_args()
 
@@ -92,9 +95,12 @@ def parse_options():
     else:
         dirname = options.directory
     res["directory"] = os.path.abspath(dirname)
+    res["directory_partial"] = os.path.abspath(dirname + "_part")
 
     if os.path.exists(res["directory"]):
         raise RuntimeError("Error: Output directory already exists: %s" % res["directory"])
+    if os.path.exists(res["directory_partial"]):
+        raise RuntimeError("Error: Partial output directory already exists: %s" % res["directory_partial"])
 
     # Verify valid --export options
     res["tables"] = []
@@ -119,13 +125,18 @@ def parse_options():
     else:
         res["fields"] = options.fields.split(",")
 
+    # Get number of clients
+    if options.clients < 1:
+       raise RuntimeError("Error: invalid number of clients (%d), must be greater than zero" % options.clients)
+    res["clients"] = options.clients
+
     res["auth_key"] = options.auth_key
     return res
 
 def get_tables(host, port, auth_key, tables):
     try:
         conn = r.connect(host, port, auth_key=auth_key)
-    except r.RqlDriverError as ex:
+    except (r.RqlError, r.RqlDriverError) as ex:
         raise RuntimeError(ex.message)
 
     dbs = r.db_list().run(conn)
@@ -148,10 +159,25 @@ def get_tables(host, port, auth_key, tables):
     # Remove duplicates by making results a set
     return set(res)
 
-def prepare_directories(base_path, db_table_set):
+def os_call_wrapper(fn, filename, error_str):
+    try:
+        fn(filename)
+    except OSError as ex:
+        raise RuntimeError(error_str % (filename, ex.strerror))
+
+# Make sure the output directory doesn't exist and create the temporary directory structure
+def prepare_directories(base_path, base_path_partial, db_table_set):
+    os_call_wrapper(lambda x: os.makedirs(x), base_path_partial, "Failed to create temporary directory (%s): %s")
+
     db_set = set([db for (db, table) in db_table_set])
     for db in db_set:
-        os.makedirs(base_path + "/%s" % db)
+        db_dir = base_path_partial + "/%s" % db
+        os_call_wrapper(lambda x: os.makedirs(x), db_dir, "Failed to create temporary directory (%s): %s")
+
+# Move the temporary directory structure over to the original output directory
+def finalize_directory(base_path, base_path_partial):
+    os_call_wrapper(lambda x: os.rename(base_path_partial, x), base_path,
+                    "Failed to move temporary directory to output directory (%s): %s")
 
 def write_table_metadata(conn, db, table, base_path):
     out = open(base_path + "/%s/%s.info" % (db, table), "w")
@@ -159,12 +185,19 @@ def write_table_metadata(conn, db, table, base_path):
     out.write(json.dumps(table_info) + "\n")
     out.close()
 
-def read_table_into_queue(conn, db, table, task_queue, exit_event):
+def read_table_into_queue(conn, db, table, task_queue, progress_info, exit_event):
+    read_rows = 0
     for row in r.db(db).table(table).run(conn, time_format="raw"):
         if exit_event.is_set():
             break
         task_queue.put([row])
 
+        # Update the progress every 20 rows - to reduce locking overhead
+        read_rows += 1
+        if read_rows % 20 == 0:
+            progress_info[0].value += 20
+    progress_info[0].value += read_rows % 20
+
 def json_writer(filename, fields, task_queue, error_queue):
     try:
         with open(filename, "w") as out:
@@ -217,34 +250,42 @@ def csv_writer(filename, fields, task_queue, error_queue):
         ex_type, ex_class, tb = sys.exc_info()
         error_queue.put((ex_type, ex_class, traceback.extract_tb(tb)))
 
-def export_table(host, port, auth_key, db, table, directory, fields, format, error_queue, exit_event):
-    task_queue = multiprocessing.queues.SimpleQueue()
-
+def launch_writer(format, directory, db, table, fields, task_queue, error_queue):
     if format == "json":
         filename = directory + "/%s/%s.json" % (db, table)
-        writer = multiprocessing.Process(target=json_writer,
-                                         args=(filename, fields, task_queue, error_queue))
+        return multiprocessing.Process(target=json_writer,
+                                       args=(filename, fields, task_queue, error_queue))
     elif format == "csv":
         filename = directory + "/%s/%s.csv" % (db, table)
-        writer = multiprocessing.Process(target=csv_writer,
-                                         args=(filename, fields, task_queue, error_queue))
+        return multiprocessing.Process(target=csv_writer,
+                                       args=(filename, fields, task_queue, error_queue))
     else:
-        error_queue.put((RuntimeError, RuntimeError("unknown format type: %s" % format),
-                         traceback.extract_tb(sys.exc_info()[2])))
-        return
+        raise RuntimeError("unknown format type: %s" % format)
+
+def export_table(host, port, auth_key, db, table, directory, fields, format, error_queue, progress_info, stream_semaphore, exit_event):
+    writer = None
 
     try:
         conn = r.connect(host, port, auth_key=auth_key)
+
+        table_size = r.db(db).table(table).count().run(conn)
+        progress_info[1].value = table_size
+        progress_info[0].value = 0
         write_table_metadata(conn, db, table, directory)
-        writer.start()
-        read_table_into_queue(conn, db, table, task_queue, exit_event)
-    except (r.RqlClientError, r.RqlDriverError) as ex:
+
+        with stream_semaphore:
+            task_queue = multiprocessing.queues.SimpleQueue()
+            writer = launch_writer(format, directory, db, table, fields, task_queue, error_queue)
+            writer.start()
+
+            read_table_into_queue(conn, db, table, task_queue, progress_info, exit_event)
+    except (r.RqlError, r.RqlDriverError) as ex:
         error_queue.put((RuntimeError, RuntimeError(ex.message), traceback.extract_tb(sys.exc_info()[2])))
     except:
         ex_type, ex_class, tb = sys.exc_info()
         error_queue.put((ex_type, ex_class, traceback.extract_tb(tb)))
     finally:
-        if writer.is_alive():
+        if writer is not None and writer.is_alive():
             task_queue.put(("exit", "event")) # Exit is triggered by sending a message with two objects
             writer.join()
         else:
@@ -255,26 +296,60 @@ def abort_export(signum, frame, exit_event, interrupt_event):
     interrupt_event.set()
     exit_event.set()
 
+def print_progress(ratio):
+    total_width = 40
+    done_width = int(ratio * total_width)
+    undone_width = total_width - done_width
+    print "\r[%s%s] %3d%%" % ("=" * done_width, " " * undone_width, int(100 * ratio)),
+    sys.stdout.flush()
+
+# We sum up the row count from all tables for total percentage completion
+#  This is because table exports can be staggered when there are not enough clients
+#  to export all of them at once.  As a result, the progress bar will not necessarily
+#  move at the same rate for different tables.
+def update_progress(progress_info):
+    rows_done = 0
+    total_rows = 1
+    for (current, max_count) in progress_info:
+        curr_val = current.value
+        max_val = max_count.value
+        if curr_val < 0:
+            # There is a table that hasn't finished counting yet, we can't report progress
+            rows_done = 0
+            break
+        else:
+            rows_done += curr_val
+            total_rows += max_val
+
+    print_progress(float(rows_done) / total_rows)
+
 def run_clients(options, db_table_set):
     # Spawn one client for each db.table
     exit_event = multiprocessing.Event()
     processes = []
     error_queue = multiprocessing.queues.SimpleQueue()
     interrupt_event = multiprocessing.Event()
+    stream_semaphore = multiprocessing.BoundedSemaphore(options["clients"])
 
     signal.signal(signal.SIGINT, lambda a,b: abort_export(a, b, exit_event, interrupt_event))
 
     try:
+        progress_info = [ ]
+
         for (db, table) in db_table_set:
+            progress_info.append((multiprocessing.Value(ctypes.c_longlong, -1),
+                                  multiprocessing.Value(ctypes.c_longlong, 0)))
             processes.append(multiprocessing.Process(target=export_table,
                                                      args=(options["host"],
                                                            options["port"],
                                                            options["auth_key"],
                                                            db, table,
-                                                           options["directory"],
+                                                           options["directory_partial"],
                                                            options["fields"],
                                                            options["format"],
                                                            error_queue,
+                                                           progress_info[-1],
+                                                           stream_semaphore,
                                                            exit_event)))
             processes[-1].start()
 
@@ -284,6 +359,15 @@ def run_clients(options, db_table_set):
             if not error_queue.empty():
                 exit_event.set() # Stop rather immediately if an error occurs
             processes = [process for process in processes if process.is_alive()]
+            update_progress(progress_info)
+
+        # If we were successful, make sure 100% progress is reported
+        # (rows could have been deleted which would result in being done at less than 100%)
+        if error_queue.empty() and not interrupt_event.is_set():
+            print_progress(1.0)
+
+        # Continue past the progress output line
+        print ""
     finally:
         signal.signal(signal.SIGINT, signal.SIG_DFL)
 
@@ -309,9 +393,14 @@ def main():
     try:
         db_table_set = get_tables(options["host"], options["port"], options["auth_key"], options["tables"])
         del options["tables"] # This is not needed anymore, db_table_set is more useful
-        prepare_directories(options["directory"], db_table_set)
+
+        # Determine the actual number of client processes we'll have
+        options["clients"] = min(options["clients"], len(db_table_set))
+
+        prepare_directories(options["directory"], options["directory_partial"], db_table_set)
         start_time = time.time()
         run_clients(options, db_table_set)
+        finalize_directory(options["directory"], options["directory_partial"])
     except RuntimeError as ex:
         print >> sys.stderr, ex
         return 1
